import os
import signal
import subprocess
import time
import tkinter
import tkinter.messagebox
import tkinter.filedialog
import uuid
from enum import Enum
from pathlib import Path

import numpy as np
from loguru import logger

from backend import load_map_from_ply_and_txt
from controllers.ControllerRosTopics import ControllerRosTopics
from controllers.ControllerMap import ControllerMap
from models.FlightPlan import FlightPlan
from models.LogRosTopic import LogRosTopic
from models.LogSeverity import LogSeverity
from models.LogRecord import LogRecord
from models.Logs import Logs
from models.PointOfInterest import PointOfInterest
from models.Project import Project
import eel

from models.RosTopics import RosTopics
from models.SegmentWall import SegmentWall
import base64

from models.StateADC import StateADC
from models.ViewSettings import ViewSettings


path_current_dir = os.path.dirname(__file__)
path_base = f'{path_current_dir}/../'

class EnumDirectoryMode(Enum):
    MODE_PLY_TXT = 'MODE_PLY_TXT'
    MODE_PROJECT = 'MODE_PROJECT'

class ControllerChromium:
    __instance = None

    def __init__(self):
        if ControllerChromium.__instance is not None:
            raise Exception("Cannot have 2 instances of singleton ControllerChromium")

        self.project = Project(
            uuid=str(uuid.uuid4())
        )
        self.filepath_txt = ''
        self.filepath_ply = ''
        self.filepath_project = ''

        eel.init(path_base + 'views', allowed_extensions=['.html'], js_result_timeout=1e9)

        self.root = tkinter.Tk()
        self.screen_width, self.screen_height = self.root.winfo_screenwidth(), self.root.winfo_screenheight()
        self.screen_height -= 50 # taskbar

        self.root.overrideredirect(True)
        self.root.wm_attributes("-topmost", "true")
        self.root.withdraw()

        self.state_adc = StateADC.NotSet
        self.listened_ros_topic_paths = {}

        # throttling functions
        self.time_last_update_state_logs = 0
        self.json_last_update_state_logs = ''
        self.time_last_update_state_frontend = 0
        self.json_last_update_state_frontend = ''
        self.time_last_set_drone_position = 0
        self.dict_last_set_drone_position = None

        self.is_recording_screen = False
        self.recording_process = None
        self.recording_filename = None

        # root.deiconify()
        # root.overrideredirect(False)
        # root.destroy()

    @staticmethod
    def instance():
        if ControllerChromium.__instance is None:
            ControllerChromium.__instance = ControllerChromium()
        return ControllerChromium.__instance

    def hide_tk_root(self):
        self.root.wm_attributes("-topmost", "false")
        self.root.overrideredirect(False)
        self.root.deiconify()
        self.root.update()
        self.root.overrideredirect(True)
        self.root.wm_attributes("-topmost", "true")
        self.root.withdraw()
        self.root.update()

    def get_default_directory(self, mode: EnumDirectoryMode):
        path_result = path_home = os.path.expanduser('~')
        path_inital = f'{path_home}'
        # path_inital = r'C:\Users\vecin\Documents\PycharmProjects\asya\alarm_flight_ui_2022_Q1\test_data'
        if os.path.exists(path_inital):
            path_result = path_inital
        if mode == EnumDirectoryMode.MODE_PROJECT:
            if self.filepath_project != '':
                path_result = Path(self.filepath_project).parent
        else:
            if self.filepath_ply != '':
                path_result = Path(self.filepath_ply).parent
            if self.filepath_txt != '':
                path_result = Path(self.filepath_txt).parent
        path_result = str(path_result)
        return path_result

    def start_executing_flight_plan(self):
        self.project.map_layers.layer_pois.current_loop_count = 0
        self.project.map_layers.layer_pois.current_delay_before_execution_sec = 0

    def update_state_frontend(self):
        json_project = self.project.to_json()
        if json_project != self.json_last_update_state_frontend or \
            time.time() - self.time_last_update_state_frontend > 0.1:
            self.json_last_update_state_frontend = json_project
            self.time_last_update_state_frontend = time.time()
            eel.update_state_frontend(json_project)
        eel.recording_status(self.is_recording_screen)

    def update_state_logs(self):
        if time.time() - self.time_last_update_state_logs > 0.1:
            # clone
            logs = Logs.from_dict(self.project.logs.to_dict())

            # sanitize
            logs.logs_ros_topics_full = None
            for log in logs.logs_ros_topics:
                log.buff_base64 = None
                log.header = None

            json_logs = logs.to_json()
            self.json_last_update_state_logs = json_logs
            self.time_last_update_state_logs = time.time()

            eel.update_state_logs(json_logs)

    def update_state_backend(self, project_json):
        self.project = Project.from_json(project_json)

    def show_message_box(self, title, message):
        self.hide_tk_root()
        tkinter.messagebox.showinfo(title, message)
        self.hide_tk_root()

    def open_filepath_txt(self):
        self.hide_tk_root()
        self.filepath_txt = tkinter.filedialog.askopenfilename(
            parent=self.root,
            title='Open a txt file',
            initialdir=self.get_default_directory(EnumDirectoryMode.MODE_PLY_TXT),
            filetypes=(("txt files", "*.txt"), ("All files", "*.*")))
        self.hide_tk_root()
        return os.path.basename(self.filepath_txt)

    def open_filepath_ply(self):
        self.hide_tk_root()
        self.filepath_ply = tkinter.filedialog.askopenfilename(
            parent=self.root,
            title='Open a PLY file',
            initialdir=self.get_default_directory(EnumDirectoryMode.MODE_PLY_TXT),
            filetypes=(("PLY files", "*.ply"), ("All files", "*.*")))
        self.hide_tk_root()
        return os.path.basename(self.filepath_ply)

    def save_project(self):
        result = ''
        try:
            if self.filepath_project == '' or not os.path.exists(self.filepath_project):
                result = self.save_as_project()
            if self.filepath_project != '':
                result = self.save_project_internal()
        except Exception as e:
            result = str(e)
        return result

    def save_project_internal(self):
        result = ''
        try:
            if len(self.filepath_project):
                project_json = self.project.to_json()
                project_base64 = base64.b64encode(project_json.encode('ascii'))
                with open(self.filepath_project, 'wb') as fp:
                    fp.write(project_base64)
                eel.set_application_filename(os.path.basename(self.filepath_project))
        except Exception as e:
            result = str(e)
        return result

    def save_as_project(self):
        result = ''
        try:
            self.hide_tk_root()
            self.filepath_project = tkinter.filedialog.asksaveasfilename(
                parent=self.root,
                title='Save flight project files',
                initialdir=self.get_default_directory(EnumDirectoryMode.MODE_PROJECT),
                filetypes=(("All files", "*.*"), ("Flight project files", "*.flight"),))
            if not self.filepath_project.endswith('.flight'):
                self.filepath_project += '.flight'
            self.hide_tk_root()
            result = self.save_project_internal()
            eel.set_application_filename(os.path.basename(self.filepath_project))
        except Exception as e:
            result = str(e)

        logger.info(f"save_project result: {result}")
        return result

    def update_view_settings_json(self, view_settings_json):
        self.project.view_settings = ViewSettings.from_json(view_settings_json)

    def update_ros_topics_json(self, ros_topics_json):
        self.project.ros_topics = RosTopics.from_json(ros_topics_json)
        self.time_last_update_state_logs = 0
        self.update_state_frontend()

    def update_flight_plan_dict(self, flight_plan_json):
        self.project.flight_plan = FlightPlan.from_dict(flight_plan_json)

    def update_pois_repeat_count(self, repeat_count):
        self.project.map_layers.layer_pois.repeat_count = repeat_count

    def update_pois_delay_before_execution_sec(self, delay_sec):
        self.project.map_layers.layer_pois.delay_before_execution_sec = delay_sec

    def open_project(self):
        result = ''
        try:
            self.hide_tk_root()
            self.filepath_project = tkinter.filedialog.askopenfilename(
                parent=self.root,
                title='Open flight project files',
                initialdir=self.get_default_directory(EnumDirectoryMode.MODE_PROJECT),
                filetypes=(("Flight project files", "*.flight"),))
            self.hide_tk_root()
            if len(self.filepath_project):
                with open(self.filepath_project, 'rb') as fp:
                    project_base64 = fp.read()
                project_json = base64.b64decode(project_base64.decode('ascii'))
                self.project = Project.from_json(project_json)
                self.update_state_frontend()
                eel.open_new_file()
                eel.set_application_filename(os.path.basename(self.filepath_project))

            self.time_last_update_state_logs = 0
            self.time_last_update_state_frontend = 0
            self.time_last_update_state_backend = 0
        except Exception as e:
            result = str(e)

        logger.info(f"open_project result: {result}")
        return result

    def convert_x_y_to_original(self, x, y):
        result = np.matmul(np.array(self.project.map_layers.transform_to_original), np.array([x, y, 1]))
        return result[0], result[1]

    def convert_x_y_from_original(self, x, y):
        result = np.matmul(np.array(self.project.map_layers.transform_from_original), np.array([x, y, 1]))
        return result[0], result[1]

    def change_poi(self, uuid, x, y):
        try:
            if self.project:
                for poi in self.project.map_layers.layer_pois.pois:
                    poi.is_selected = False

            x, y = self.convert_x_y_to_original(x, y)
            if uuid == "":
                self.add_log_info(
                    'UI',
                    f'Set position to {round(x, 2)}, {round(y, 2)}'
                )
            else:
                poi = self.get_poi_by_uuid(uuid)
                if poi:
                    poi.is_selected = True
                    self.add_log_info(
                        'UI',
                        f'Selected {poi.name} at position {round(x, 2)}, {round(y, 2)}'
                    )
                    eel.poi_set_selection(poi.uuid);
        except Exception as e:
            self.add_log_error(e.message)

    def set_state_adc(self, state: StateADC):
        self.state_adc = state
        eel.set_state_adc(state.value)

    def save_poi(self, poi_name, x, y, poi_uuid):
        if poi_uuid == "":
            poi = PointOfInterest(
                uuid=str(uuid.uuid4()),
                name=poi_name,
                x=x,
                y=y
            )
            self.add_log_info('UI', f'Save POI: {poi_name} {round(x, 2)}, {round(y, 2)}')
            self.project.map_layers.layer_pois.pois.append(poi)
        else:
            poi = self.get_poi_by_uuid(poi_uuid)
            self.add_log_info('UI', f'Rename POI: {poi.name} to {poi_name}')
            if poi:
                poi.name = poi_name
        self.update_state_frontend()

    def rename_poi(self, poi_name_before, poi_name_after):
        self.add_log_info('UI', f'Rename POI: {poi_name_before}')
        pois = []
        for poi in self.project.map_layers.layer_pois.pois:
            if poi.name == poi_name_before:
                poi.name = poi_name_after
            pois.append(poi)
        self.project.map_layers.layer_pois.pois = pois
        self.update_state_frontend()

    def change_pois_order(self, uuids_order):
        self.add_log_info('UI', f'Changing order of POI')
        pois = []
        for uuid in uuids_order:
            pois.append(self.get_poi_by_uuid(uuid))
        self.project.map_layers.layer_pois.pois = pois
        self.update_state_frontend()

    def get_poi_by_uuid(self, uuid):
        result = None
        for poi in self.project.map_layers.layer_pois.pois:
            if poi.uuid == uuid:
                result = poi
                break
        return result

    def delete_poi(self, uuid):
        poi = self.get_poi_by_uuid(uuid)
        if poi:
            self.add_log_info('UI', f'Delete POI: {poi.name}')
        self.project.map_layers.layer_pois.pois = [it for it in self.project.map_layers.layer_pois.pois if it.uuid != uuid]
        self.update_state_frontend()

    def clear_all_poi(self):
        self.add_log_info('UI', 'Delete all POIs')
        self.project.map_layers.layer_pois.pois = []
        self.update_state_frontend()

    def set_drone_position(self, x_orig, y_orig, z_orig, dir_x_orig, dir_y_orig, dir_z_orig, dir_w_orig):
        x, y = self.convert_x_y_from_original(x_orig, y_orig)
        angle = 0
        dir_x = 0
        dir_y = 0

        try:
            dir = np.array([dir_x_orig, dir_y_orig])
            dir_norm = np.linalg.norm(dir, ord=2)
            dir = dir / dir_norm
            dir_static = np.array([0, 1])

            angle = np.degrees(np.arctan2(dir_static[1], dir_static[0]) - np.arctan2(dir[1], dir[0]))
            angle -= 180

            dir_x = dir[0]
            dir_y = dir[1]
        except Exception as e:
            logger.exception(e)
        if time.time() - self.time_last_set_drone_position > 0.05:
            self.time_last_set_drone_position = time.time()

            if self.dict_last_set_drone_position is None or \
                (self.dict_last_set_drone_position["x"] != x or
                    self.dict_last_set_drone_position["y"] != y or
                    self.dict_last_set_drone_position["angle"] != angle):

                self.dict_last_set_drone_position = {
                    "x": x,
                    "y": y,
                    "angle": angle
                }
                eel.set_drone_position(
                    x, y, dir_x, dir_y, angle,
                    x_orig, y_orig
                )

    def clear_logs(self):
        self.project.logs = Logs()
        self.update_state_frontend()

    def add_log_ros_topic(
            self,
            topic_full_path: str,
            message_type: str,
            message: str,
            buff: str,
            header: dict,
    ):
        timestamp_unix = int(time.time())
        new_record = LogRosTopic(
            timestamp_unix=timestamp_unix,
            topic_full_path=topic_full_path,
            message_type=message_type,
            message=message,
            buff_base64=buff,
            header=header
        )
        self.project.logs.logs_ros_topics_full.append(new_record)
        if self.listened_ros_topic_paths.get(topic_full_path):
            self.project.logs.logs_ros_topics.append(new_record)
            self.update_state_logs()

    def update_ros_log_filter(self):
        for ros_topic in ControllerRosTopics.get_ros_topics_flattened(self.project.ros_topics):
            if len(ros_topic.children_topics) == 0:
                self.listened_ros_topic_paths[ros_topic.full_path] = ros_topic.is_selected

        new_logs = []
        for log in self.project.logs.logs_ros_topics_full:
            if self.listened_ros_topic_paths.get(log.topic_full_path):
                new_logs.append(log)
        self.project.logs.logs_ros_topics = new_logs
        self.update_state_logs()

    def add_log(
            self,
            timestamp_unix: int,
            source: str,
            severity: LogSeverity,
            message: str
    ):
        new_record = LogRecord(
            timestamp_unix,
            source,
            severity,
            message
        )
        if len(self.project.logs.logs) == 0 or new_record.message != self.project.logs.logs[-1].message:
            logger.info(f'{source} {message}')
            self.project.logs.logs.append(new_record)
            self.update_state_logs()

    def add_log_info(
            self,
            source: str,
            message: str
    ):
        severity = LogSeverity.INFO
        timestamp_unix = int(time.time())
        self.add_log(
            timestamp_unix,
            source,
            severity,
            message
        )

    def add_log_error(
            self,
            source: str,
            message: str
    ):
        severity = LogSeverity.ERROR
        timestamp_unix = int(time.time())
        self.add_log(
            timestamp_unix,
            source,
            severity,
            message
        )

    @staticmethod
    def set_loading_progress(progress, message):
        eel.set_loading_progress(progress, message)

    def load_ply_and_txt(self):
        logger.info("load_ply_and_txt start")

        result = ''
        try:
            ControllerChromium.set_loading_progress(0, 'Loading ply and txt')
            self.project = Project(
                uuid=str(uuid.uuid4())
            )
            project = self.project
            controllerMap = ControllerMap(self.filepath_ply, self.filepath_txt, use_45deg_lines=False)

            ControllerChromium.set_loading_progress(85, 'Scaling map')
            project.map_layers.transform_to_original = \
                controllerMap.mapTemporary.transformation_matrix_pix_to_meters.tolist()
            project.map_layers.transform_from_original = \
                controllerMap.mapTemporary.transformation_matrix_meters_to_pix.tolist()

            ControllerChromium.set_loading_progress(90, 'Parsing points for rendering')
            project.map_layers.layer_pointcloud_raw.x = \
                [round(it.x, 2) for it in controllerMap.mapTemporary.point_cloud_in_pix.points]
            project.map_layers.layer_pointcloud_raw.y = \
                [round(it.y, 2) for it in controllerMap.mapTemporary.point_cloud_in_pix.points]

            project.map_layers.layer_pointcloud_filtered.x = \
                [round(it.x, 2) for it in controllerMap.mapTemporary.filtered_point_cloud_in_pix.points]
            project.map_layers.layer_pointcloud_filtered.y = \
                [round(it.y, 2) for it in controllerMap.mapTemporary.filtered_point_cloud_in_pix.points]

            project.map_layers.layer_drone_path.x = \
                [round(it.x, 2) for it in controllerMap.mapTemporary.path_points_in_pix.points]
            project.map_layers.layer_drone_path.y = \
                [round(it.y, 2) for it in controllerMap.mapTemporary.path_points_in_pix.points]

            ControllerChromium.set_loading_progress(95, 'Parsing walls for rendering')
            for idx in range(len(controllerMap.mapTemporary.line_start_points_in_pix)):
                segment_wall = SegmentWall()

                segment_wall.x.append(round(controllerMap.mapTemporary.line_start_points_in_pix[idx, 0], 2))
                segment_wall.x.append(round(controllerMap.mapTemporary.line_end_points_in_pix[idx, 0], 2))

                segment_wall.y.append(round(controllerMap.mapTemporary.line_start_points_in_pix[idx, 1], 2))
                segment_wall.y.append(round(controllerMap.mapTemporary.line_end_points_in_pix[idx, 1], 2))

                project.map_layers.layer_walls.walls.append(segment_wall)

            ControllerChromium.set_loading_progress(100, 'Done')
            self.update_state_frontend()
            eel.open_new_file()
            eel.set_application_filename(os.path.basename(self.filepath_ply))

            self.time_last_update_state_logs = 0
            self.time_last_update_state_frontend = 0
            self.time_last_update_state_backend = 0
        except Exception as e:
            result = str(e)
            logger.exception(result)

        logger.info(f"load_ply_and_txt result: {result}")
        return result

    def update_ros_topics(self):
        ros_topics_before = ControllerRosTopics.get_ros_topics_flattened(self.project.ros_topics)
        available_ros_topics = ControllerRosTopics.get_ros_topics_flattened(ControllerRosTopics.instance().get_ros_topics())

        ros_topics_merged = list(ros_topics_before)
        for topic in available_ros_topics:
            is_topic_before = False
            for topic_before in ros_topics_before:
                if topic_before.full_path == topic.full_path:
                    is_topic_before = True
                    break
            if not is_topic_before:
                ros_topics_merged.append(topic)

        ros_topics_merged = sorted(ros_topics_merged, key=lambda x: x.full_path)
        self.project.ros_topics = ControllerRosTopics.get_ros_topics_tree(ros_topics_merged)
        self.update_state_frontend()

    def write_to_bag_file(self):
        self.hide_tk_root()
        path = tkinter.filedialog.asksaveasfilename(
            parent=self.root,
            title='Save bag files',
            initialdir=self.get_default_directory(EnumDirectoryMode.MODE_PROJECT),
            filetypes=(("All files", "*.*"), ("Bag files", "*.bag"),))
        if isinstance(path, str) and len(path) > 0:
            if not path.endswith('.bag'):
                path += '.bag'
            self.hide_tk_root()
            ControllerRosTopics.instance().write_to_bag_file(self.project.logs.logs_ros_topics, path)

    def start_recording_screen(self):
        try:
            self.recording_filename = self.get_default_directory(EnumDirectoryMode.MODE_PROJECT)
            self.recording_filename += f"/alarm_screen_recording_{round(time.time())}.mkv"
            self.recording_process = subprocess.Popen([
                "ffmpeg",
                "-video_size", f"{self.screen_width}x{self.screen_height + 50}",
                "-f", "x11grab",
                "-i", ":0.0",
                "-framerate", "25",
                "-preset", "ultrafast",
                self.recording_filename
            ])
            self.add_log_info("REC", "Screen recording started. Filename: " + self.recording_filename)
            self.is_recording_screen = True
        except Exception as e:
            logger.exception(e)
        finally:
            eel.recording_status(self.is_recording_screen)

    def stop_recording(self):
        try:
            if self.recording_process is not None:
                self.recording_process.send_signal(signal.SIGINT)
                self.recording_process = None
                self.add_log_info("REC", "Screen recording stopped. File saved to: " + self.recording_filename)
        except Exception as e:
            logger.exception(e)
        self.is_recording_screen = False
        self.recording_filename = None
        eel.recording_status(self.is_recording_screen)
