import copy

import numpy as np

from models.Line import Line
from models.Point import Point
from controllers.ControllerPoint import ControllerPoint

from shapely.geometry.linestring import LineString

# TODO: ControllerPoint must have transform, to_list, distance functions
class ControllerLine:

    @staticmethod
    def create_line(p1, p2, angle=None):
        if isinstance(p1, Point):
            start_point = p1
        else:
            start_point = Point(p1[0], p1[1])
        if isinstance(p2, Point):
            end_point = p2
        else:
            end_point = Point(p2[0], p2[1])

        # order points so start closer to 0,0
        p1_dist = np.sqrt((start_point.x - 0)**2 + (start_point.y - 0)**2)
        p2_dist = np.sqrt((end_point.x - 0)**2 + (end_point.y - 0)**2)
        if p2_dist < p1_dist:
            tmp_p = start_point
            start_point = end_point
            end_point = tmp_p

        if angle is None:
            angle = np.arctan2(end_point.y - start_point.y, end_point.x - start_point.x)
        else:
            angle = angle
        line = Line(start_point=start_point, end_point=end_point, angle=angle)
        return line

    @staticmethod
    def delta_angle(line: Line, angle):
        return np.atan2(np.sin(angle - line.angle), np.cos(angle - line.angle))

    @staticmethod
    def intersects_line(line_1: Line, line_2: Line):
        linestring_1 = ControllerLine.to_linestring(line_1)
        linestring_2 = ControllerLine.to_linestring(line_2)
        return linestring_1.intersection(linestring_2) if linestring_1.intersects(linestring_2) else False

    @staticmethod
    def transform(line: Line, transform):
        new_start_point = ControllerPoint.transform(line.start_point, transform)
        new_end_point = ControllerPoint.transform(line.end_point, transform)
        return ControllerLine.create_line(new_start_point, new_end_point)

    @staticmethod
    def extend_line(line: Line, length, extent_end):
        tmp_line=copy.deepcopy(line)
        if extent_end:
            tmp_line.end_point.x = tmp_line.end_point.x + length * np.cos(tmp_line.angle)
            tmp_line.end_point.y = tmp_line.end_point.y + length * np.sin(tmp_line.angle)
        else:
            tmp_line.start_point.x = tmp_line.start_point.x - length * np.cos(tmp_line.angle)
            tmp_line.start_point.y = tmp_line.start_point.y - length * np.sin(tmp_line.angle)
        return tmp_line

    @staticmethod
    def shorten_line(line: Line, length, cut_end):
        tmp_line = copy.deepcopy(line)
        if length > ControllerLine.get_lenght(line):
            return None
        if cut_end:
            tmp_line.end_point.x = tmp_line.end_point.x - length * np.cos(tmp_line.angle)
            tmp_line.end_point.y = tmp_line.end_point.y - length * np.sin(tmp_line.angle)
        else:
            tmp_line.start_point.x = tmp_line.start_point.x + length * np.cos(tmp_line.angle)
            tmp_line.start_point.y = tmp_line.start_point.y + length * np.sin(tmp_line.angle)
        return tmp_line

    @staticmethod
    def rotate_line(line: Line, angle):
        # absolute angle
        tmp_line = copy.deepcopy(line)
        len_from_midpoint = ControllerPoint.distance(ControllerLine.midpoint(line), line.start_point)
        tmp_line.start_point.x = ControllerLine.midpoint(line).x - len_from_midpoint * np.cos(angle)
        tmp_line.start_point.y = ControllerLine.midpoint(line).y - len_from_midpoint * np.sin(angle)
        tmp_line.end_point.x = ControllerLine.midpoint(line).x + len_from_midpoint * np.cos(angle)
        tmp_line.end_point.y = ControllerLine.midpoint(line).y + len_from_midpoint * np.sin(angle)
        tmp_line.angle = angle
        return tmp_line

    @staticmethod
    def side_shift_line(line: Line, distance):
        # absolute angle
        tmp_line = copy.deepcopy(line)
        tmp_line.start_point.x = line.start_point.x + distance * np.cos(line.angle + np.pi/2)
        tmp_line.start_point.y = line.start_point.y + distance * np.sin(line.angle + np.pi/2)
        tmp_line.end_point.x = line.end_point.x + distance * np.cos(line.angle + np.pi/2)
        tmp_line.end_point.y = line.end_point.y + distance * np.sin(line.angle + np.pi/2)
        return tmp_line

    @staticmethod
    def to_list(line: Line):
        return [*ControllerPoint.to_list(line.start_point), *ControllerPoint.to_list(line.end_point)]

    @staticmethod
    def midpoint(line: Line):
        return Point(x=(line.start_point.x + line.end_point.x) / 2, y=(line.start_point.y + line.end_point.y) / 2)

    @staticmethod
    def get_lenght(line: Line):
        return np.sqrt((line.start_point.x - line.end_point.x)**2 + (line.start_point.y - line.end_point.y)**2)

    @staticmethod
    def angle_as_sin_cos(line: Line):
        return np.sin(line.angle), np.cos(line.angle)

    @staticmethod
    def to_linestring(line: Line):
        return LineString((ControllerPoint.to_list(line.start_point), ControllerPoint.to_list(line.end_point)))

    @staticmethod
    def from_linestring(line: LineString):
        coords = list(line.coords)
        return ControllerLine.create_line(coords[0], coords[1])
