import argparse
import json
import matplotlib
import os
import copy
import cv2

import numpy as np

from scipy.spatial.transform import Rotation as R
from skimage.transform import hough_line, hough_line_peaks
from scipy import ndimage
from sklearn.cluster import DBSCAN
from plyfile import PlyData  # pip install plyfile
from shapely.geometry.linestring import LineString
from shapely.ops import nearest_points, split
from loguru import logger
import matplotlib.pyplot as plt

from models.Line import Line
from models.Point import Point
from models.MapTemporary import MapTemporary
from controllers.ControllerLine import ControllerLine
from controllers.ControllerPoint import ControllerPoint


class ControllerMap:
    def __init__(self, filepath_ply=None, filepath_txt=None, use_45deg_lines=False):
        # TODO remove this when 45 degree angles work better
        # use_45deg_lines = True

        self.filepath_ply = filepath_ply
        self.filepath_txt = filepath_txt
        self.use_45deg_lines = use_45deg_lines
        self.mapTemporary = MapTemporary()
        self.controllerPoint = ControllerPoint()
        self.controllerLine = ControllerLine()

        arg_parser = argparse.ArgumentParser()
        arg_parser.add_argument('-use_matplotlib_agg', default=False, type=lambda x: (str(x).lower() == 'true'))

        arg_parser.add_argument('-debug_draw', default=False, type=lambda x: (str(x).lower() == 'true'))

        # map args
        arg_parser.add_argument('-pix_per_meter', default=50, type=int)
        arg_parser.add_argument('-map_padding', default=0.2, type=float)  # part of map size
        arg_parser.add_argument('-wall_width', default=0.2, type=float)  # in meters
        arg_parser.add_argument('-drone_width', default=0.2, type=float)  # in meters
        arg_parser.add_argument('-use_45deg_lines', default=False, type=lambda x: (
                    str(x).lower() == 'true'))  # if true will search diagonal lines else only 90 deg lines
        arg_parser.add_argument('-snap_line_angle_delta_deg', default=5, type=float) # in degrees

        # line processing args
        arg_parser.add_argument('-max_wall_gap', default=0.4, type=float)  # in meters
        arg_parser.add_argument('-line_extend_step', default=0.2, type=float)  # in meters
        arg_parser.add_argument('-min_line_distance', default=1.0, type=float)  # in meters
        arg_parser.add_argument('-min_line_length', default=0.4, type=float)  # in meters
        arg_parser.add_argument('-max_shift_after_rotate', default=0.2, type=float)  # in meters
        arg_parser.add_argument('-min_pixel_density', default=0.25, type=float)  # range 0 - 1
        arg_parser.add_argument('-min_diagonal_line_length', default=55.0, type=float)  # in pix
        arg_parser.add_argument('-parallel_join_distance', default=1.0, type=float)  # in meters

        # Hough transform params
        arg_parser.add_argument('-is_probabilistic_hough', default=True, type=lambda x: (str(x).lower() == 'true'))
        arg_parser.add_argument('-hough_rho', default=1, nargs='*', type=float)
        arg_parser.add_argument('-hough_theta', default=np.pi / 180, nargs='*', type=float)
        # higher threshold is needed for non-probabilistic Hough transform (40 for the first map)
        arg_parser.add_argument('-hough_threshold', default=10, nargs='*', type=int)
        arg_parser.add_argument('-hough_minLineLength', default=5, nargs='*', type=int)
        arg_parser.add_argument('-hough_maxLineGap', default=10, nargs='*', type=int)

        # Canny detector params
        arg_parser.add_argument('-canny_threshold1', default=50, nargs='*', type=int)
        arg_parser.add_argument('-canny_threshold2', default=200, nargs='*', type=int)
        arg_parser.add_argument('-canny_aperatureSize', default=3, nargs='*',
                                type=int)  # usually between 3 and 7 (kernel size)
        arg_parser.add_argument('-canny_L2grad', default=True, type=lambda x: (str(x).lower() == 'true'))

        # Blob detection params (default params detect single points)
        arg_parser.add_argument('-blobs_thresholdStep', default=1, nargs='*', type=int)  # don't change this
        arg_parser.add_argument('-blobs_minThreshold', default=127, nargs='*', type=int)  # don't change this
        arg_parser.add_argument('-blobs_maxThreshold', default=129, nargs='*', type=int)  # don't change this
        arg_parser.add_argument('-blobs_minDistBetweenBlobs', default=10.0, nargs='*', type=float)
        arg_parser.add_argument('-blobs_minArea', default=1, nargs='*', type=float)
        arg_parser.add_argument('-blobs_maxArea', default=10, nargs='*', type=float)
        arg_parser.add_argument('-blobs_minCircularity', default=0.7, nargs='*', type=float)
        arg_parser.add_argument('-blobs_minConvexity', default=0.7, nargs='*', type=float)
        arg_parser.add_argument('-blobs_minInertiaRatio', default=0.7, nargs='*', type=float)

        self.args, args_other = arg_parser.parse_known_args()
        logger.info('Map processing script starting')
        logger.info(json.dumps(self.args.__dict__))

        if self.args.use_matplotlib_agg:
            matplotlib.use('Agg')

        if not self.args.debug_draw:
            from controllers.ControllerChromium import ControllerChromium
            ControllerChromium.set_loading_progress(10, 'Creating map')
        self._init_map_temporary()
        if not self.args.debug_draw:
            ControllerChromium.set_loading_progress(20, 'Fixing map rotation')
        self._rotate_map()
        if not self.args.debug_draw:
            ControllerChromium.set_loading_progress(30, 'Cleaning Finding blobs of points')
        self.mapTemporary.map_image_rgb = self.detect_blobs(show=False)
        if not self.args.debug_draw:
            ControllerChromium.set_loading_progress(40, 'Creating clean point-cloud')
        self._init_filtered_point_cloud()
        if not self.args.debug_draw:
            ControllerChromium.set_loading_progress(50, 'Finding potential walls')
        temp_lines = self.detect_lines()
        if not self.args.debug_draw:
            ControllerChromium.set_loading_progress(60, 'Finding walls and adding them to the map')
        logger.info(f'Lines found {len(temp_lines)}')
        self.process_and_add_lines(temp_lines)
        if not self.args.debug_draw:
            ControllerChromium.set_loading_progress(70, 'Processing path')

        if self.filepath_txt is not None:
            self.process_path()

        if not self.args.debug_draw:
            ControllerChromium.set_loading_progress(80, 'Fixing map rotation and scale #2')
        self.rotate_and_scale_data()


    def _debug_draw(self, map, lines, title='', thinckness=1):
        tmp_image = copy.deepcopy(map)
        for idx in range(0, len(lines)):
            before_pix = np.array([lines[idx].start_point.x, lines[idx].start_point.y])
            after_pix = np.array([lines[idx].end_point.x, lines[idx].end_point.y])
            cv2.line(tmp_image, (int(before_pix[0]), int(before_pix[1])), (int(after_pix[0]), int(after_pix[1])), color=(255,0,0), thickness=thinckness)
        plt.imshow(tmp_image)
        plt.gca().invert_yaxis()
        plt.title(title)
        plt.show(block=True)

    def _init_point_cloud_from_ply(self):
        with open(self.filepath_ply, 'rb') as f:
            plydata = PlyData.read(f)

        vertices_x = plydata.elements[0].data['x']
        vertices_y = plydata.elements[0].data['y']

        vertices = np.column_stack((vertices_x, vertices_y))
        self.mapTemporary.point_cloud_in_meters.points = [Point(p[0], p[1]) for p in vertices]

    def _init_drone_path_from_txt(self):
        if os.path.isfile(self.filepath_txt):
            with open(self.filepath_txt, 'rb') as f:
                lines = f.readlines()
                for idx in range(len(lines) - 2):
                    line = lines[idx].decode('utf-8')
                    line = line.split(sep=" ")
                    self.mapTemporary.path_points_in_meters.points.append(Point(float(line[3]), float(line[7])))
                    r = R.from_matrix([[float(line[0]), float(line[1]), float(line[2])],
                                       [float(line[4]), float(line[5]), float(line[6])],
                                       [float(line[8]), float(line[9]), float(line[10])]])
                    self.mapTemporary.path_points_in_meters.w.append(r.as_quat()[3])

    def _init_map_temporary(self):
        if self.filepath_ply is not None:
            self._init_point_cloud_from_ply()
        if self.filepath_txt is not None:
            self._init_drone_path_from_txt()

        x_vertices = [p.x * self.args.pix_per_meter for p in self.mapTemporary.point_cloud_in_meters.points]
        x_delta = np.max(x_vertices) - np.min(x_vertices)
        y_vertices = [p.y * self.args.pix_per_meter for p in self.mapTemporary.point_cloud_in_meters.points]
        y_delta = np.max(y_vertices) - np.min(y_vertices)
        size = int(max(x_delta, y_delta))
        padding_pix = int(self.args.map_padding * size / 2)
        new_size = size + int(padding_pix * 2)
        if x_delta > y_delta:
            shift_x = padding_pix + abs(np.min(x_vertices))
            ratio = y_delta / x_delta
            padding_other = (new_size - (size * ratio)) / 2
            shift_y = padding_other + abs(np.min(y_vertices))
        else:
            shift_y = padding_pix + abs(np.min(y_vertices))
            ratio = x_delta / y_delta
            padding_other = (new_size - (size * ratio)) / 2
            shift_x = padding_other + abs(np.min(x_vertices))

        self.mapTemporary.transformation_matrix_meters_to_pix = np.array([
            [self.args.pix_per_meter, 0, shift_x],
            [0, self.args.pix_per_meter, shift_y],
            [0, 0, 1]
        ])
        self.mapTemporary.transformation_matrix_pix_to_meters = np.linalg.inv(self.mapTemporary.transformation_matrix_meters_to_pix)

        self.mapTemporary.map_image_rgb = np.zeros((new_size, new_size, 3), np.uint8)
        self.mapTemporary.map_image_greyscale = np.zeros((new_size, new_size), np.uint8)

        for point in self.mapTemporary.point_cloud_in_meters.points:
            vertex_pix = np.matmul(self.mapTemporary.transformation_matrix_meters_to_pix, np.array([point.x, point.y, 1]))
            self.mapTemporary.point_cloud_in_pix.points.append(Point(vertex_pix[0], vertex_pix[1]))
            cv2.circle(self.mapTemporary.map_image_rgb,
                       center=(int(vertex_pix[0]), int(vertex_pix[1])),
                       radius=0,
                       color=(255, 255, 255),
                       thickness=2)

    def _rotate_map(self):
        tested_angles = np.linspace(-np.pi / 2, np.pi / 2, 360, endpoint=False)
        image = self.mapTemporary.map_image_rgb[:, :, 0]
        h, theta, d = hough_line(image, theta=tested_angles)
        h, theta, d = hough_line_peaks(h, theta, d)
        rotate_angle = theta[np.argmax(h)] - np.pi / 2

        x = list(map(lambda point: point.x, self.mapTemporary.point_cloud_in_meters.points))
        y = list(map(lambda point: point.y, self.mapTemporary.point_cloud_in_meters.points))
        delta_x = max(x) - min(x)
        delta_y = max(y) - min(y)
        if delta_y > delta_x:
            rotate_angle += np.pi / 2

        # rotate map
        self.mapTemporary.map_image_rgb = ndimage.rotate(self.mapTemporary.map_image_rgb,
                                                         180 * rotate_angle / np.pi,
                                                         reshape=False)

        # add rotation to transformation matrix
        center_image_pix = Point(*np.array(self.mapTemporary.map_image_rgb.shape[:2]) / 2)
        # center_image_meters = center_image_pix.transform(self.mapTemporary.transformation_matrix_pix_to_meters)
        center_image_meters = self.controllerPoint.transform(center_image_pix, self.mapTemporary.transformation_matrix_pix_to_meters)
        translate_to_image = np.array([
            [1, 0, -center_image_meters.x],
            [0, 1, -center_image_meters.y],
            [0, 0, 1]
        ])
        rotate_angle = -rotate_angle
        rotate_matrix = np.array([
            [np.cos(rotate_angle), -np.sin(rotate_angle), 0],
            [np.sin(rotate_angle), np.cos(rotate_angle), 0],
            [0, 0, 1]
        ])
        translate_to_origin = np.array([
            [1, 0, center_image_meters.x],
            [0, 1, center_image_meters.y],
            [0, 0, 1]
        ])
        self.mapTemporary.transformation_matrix_meters_to_pix = np.matmul(
            np.matmul(np.matmul(self.mapTemporary.transformation_matrix_meters_to_pix, translate_to_origin),
                      rotate_matrix), translate_to_image)
        self.mapTemporary.transformation_matrix_pix_to_meters = \
            np.linalg.inv(self.mapTemporary.transformation_matrix_meters_to_pix)

    def detect_blobs(self, show=False):
        # Good visualisation of parameters:
        # https://learnopencv.com/blob-detection-using-opencv-python-c/

        # Setup SimpleBlobDetector parameters
        params = cv2.SimpleBlobDetector_Params()

        # Change thresholds
        params.thresholdStep = self.args.blobs_thresholdStep
        params.minThreshold = self.args.blobs_minThreshold
        params.maxThreshold = self.args.blobs_maxThreshold

        # Filter by color
        params.filterByColor = False

        params.minDistBetweenBlobs = self.args.blobs_minDistBetweenBlobs

        # Filter by Area.
        params.filterByArea = True
        params.minArea = self.args.blobs_minArea
        params.maxArea = self.args.blobs_maxArea

        # Filter by Circularity
        params.filterByCircularity = True
        params.minCircularity = self.args.blobs_minCircularity
        # params.maxCircularity = 1

        # Filter by Convexity
        params.filterByConvexity = True
        params.minConvexity = self.args.blobs_minConvexity
        # params.maxConvexity = 1

        # Filter by Inertia
        params.filterByInertia = True
        params.minInertiaRatio = self.args.blobs_minInertiaRatio

        # Create a detector with the parameters
        detector = cv2.SimpleBlobDetector_create(params)

        map_image_rgb = copy.copy(self.mapTemporary.map_image_rgb)
        # Detect blobs.
        keypoints = detector.detect(map_image_rgb)

        # Draw detected blobs as red circles.
        # cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS ensures the size of the circle corresponds to the size of blob
        img_with_keypoints = cv2.drawKeypoints(map_image_rgb, keypoints, np.array([]), (0, 0, 255),
                                              cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS)

        img_without_blobs = map_image_rgb
        for keypoint in keypoints:
            img_without_blobs = cv2.circle(img_without_blobs,
                                           center=(int(keypoint.pt[0]), int(keypoint.pt[1])),
                                           radius=int(keypoint.size / 2),
                                           color=(0, 0, 0),
                                           thickness=-1)

        if show:
            cv2.imshow("Keypoints", img_with_keypoints)
            cv2.waitKey(0)

            cv2.imshow("Keypoints colored out", img_without_blobs)
            cv2.waitKey(0)

            cv2.destroyAllWindows()

        return img_without_blobs

    def get_filtered_map_in_pix_unrotated(self):
        binary_img = self.mapTemporary.map_image_rgb[:, :, 0]
        filtered_map_in_pix_unrotated = cv2.findNonZero(binary_img).squeeze(1)
        return filtered_map_in_pix_unrotated

    def _init_filtered_point_cloud(self):
        # TODO: refactor this function so it is easier to read (possibly split in multiple functions and/or comment)
        filtered_map_in_pix = self.get_filtered_map_in_pix_unrotated()
        point_cloud_in_meters = copy.deepcopy(self.mapTemporary.point_cloud_in_meters.points)
        for idx in range(len(point_cloud_in_meters)):
            result = np.matmul(self.mapTemporary.transformation_matrix_meters_to_pix,
                               np.array([point_cloud_in_meters[idx].x, point_cloud_in_meters[idx].y, 1]))
            point_cloud_in_meters[idx].x = result[0]
            point_cloud_in_meters[idx].y = result[1]

        x = list(map(lambda point: point.x, point_cloud_in_meters))
        y = list(map(lambda point: point.y, point_cloud_in_meters))
        merged_list = [[round(x[i]), round(y[i])] for i in range(0, len(x))]
        merged_ndarray = np.array(merged_list)

        filtered_map_in_pix = filtered_map_in_pix.astype(merged_ndarray.dtype)

        dtype_for_intersect = {'names': ['field_0', 'field_1'], 'formats': 2 * [merged_ndarray.dtype]}

        values, idx_a, idx_b = np.intersect1d(merged_ndarray.view(dtype_for_intersect),
                                              filtered_map_in_pix.view(dtype_for_intersect),
                                              return_indices=True)

        point_cloud_in_meters = copy.deepcopy(self.mapTemporary.point_cloud_in_meters.points)
        self.mapTemporary.filtered_point_cloud_in_meters.points = [point_cloud_in_meters[i] for i in idx_a]

        clustering = DBSCAN(min_samples=3, eps=0.1) # TODO review this clustering; test on multiple OS
        x = list(map(lambda point: point.x, self.mapTemporary.filtered_point_cloud_in_meters.points))
        y = list(map(lambda point: point.y, self.mapTemporary.filtered_point_cloud_in_meters.points))
        merged_list = [[x[i], y[i]] for i in range(0, len(x))]
        result = clustering.fit_predict(merged_list)
        point_cloud = copy.deepcopy(self.mapTemporary.filtered_point_cloud_in_meters.points)
        self.mapTemporary.filtered_point_cloud_in_meters.points.clear()
        for i in range(len(result)):
            if result[i] != -1:
                self.mapTemporary.filtered_point_cloud_in_meters.points.append(point_cloud[i])

    def detect_lines(self):
        # TODO: remove the unnecessary case:
        if self.mapTemporary.map_image_rgb.ndim == 2:
            blackAndWhiteImage = self.mapTemporary.map_image_rgb
        else:
            # Convert the image to gray-scale
            grayImage = cv2.cvtColor(self.mapTemporary.map_image_rgb, cv2.COLOR_BGR2GRAY)
            (thresh, blackAndWhiteImage) = cv2.threshold(grayImage, 127, 255, cv2.THRESH_BINARY)

        # Find the edges in the image using canny detector
        edges = cv2.Canny(blackAndWhiteImage,
                          threshold1=self.args.canny_threshold1,
                          threshold2=self.args.canny_threshold1,
                          apertureSize=self.args.canny_aperatureSize,
                          L2gradient=self.args.canny_L2grad)
        # Detect points that form a line
        if self.args.is_probabilistic_hough:
            lines = cv2.HoughLinesP(edges,
                                    rho=self.args.hough_rho,
                                    theta=self.args.hough_theta,
                                    threshold=self.args.hough_threshold,
                                    minLineLength=self.args.hough_minLineLength,
                                    maxLineGap=self.args.hough_maxLineGap)
            lines_obj = []
            for line in lines:
                start_point = Point(line[0, 0], line[0, 1])
                end_point = Point(line[0, 2], line[0, 3])
                lines_obj.append(ControllerLine.create_line(start_point, end_point))
        else:
            lines = cv2.HoughLines(edges,
                                   rho=self.args.hough_rho,
                                   theta=self.args.hough_theta,
                                   threshold=self.args.hough_threshold)
            lines_obj = []
            for line in lines:
                for rho, theta in line:
                    a = np.cos(theta)
                    b = np.sin(theta)
                    x0 = a * rho
                    y0 = b * rho
                    x1 = int(x0 + 1000 * (-b))
                    y1 = int(y0 + 1000 * (a))
                    x2 = int(x0 - 1000 * (-b))
                    y2 = int(y0 - 1000 * (a))
                    start_point = Point(x1, y1)
                    end_point = Point(x2, y2)
                    lines_obj.append(ControllerLine.create_line(start_point, end_point))

        return lines_obj

    def process_and_add_lines(self, lines):
        '''
        This method processes all hough lines to lines that are aligned with walls.
        Steps:
        1. Correcting line angles to match to closest 45 or 90 angle.
        2. Side shift line to get maximum pixel coverage.
        3. Extend line length till no new pixels covered.
        4. Check if line is not too close final lines.
        5. Check pixel density of the line.
        6. Check if line crosses already added lines.
        7. if passes 4., 5. and 6. add to final lines.
        '''
        # lines_lengths = [l.transform(self.mapTemporary.transformation_matrix_pix_to_meters).length for l in lines]
        lines_lengths = [self.controllerLine.get_lenght(self.controllerLine.transform(l, self.mapTemporary.transformation_matrix_pix_to_meters)) for l in lines]
        # order descending
        lines = list(np.array(lines)[np.argsort(lines_lengths)[::-1]])
        lines_lengths = [self.controllerLine.get_lenght(self.controllerLine.transform(l, self.mapTemporary.transformation_matrix_pix_to_meters)) for l in lines]
        lines = list(np.array(lines)[np.where(np.array(lines_lengths) > self.args.min_line_length)])
        logger.info(f'Lines found {len(lines)}')

        map_for_coverage = np.array(np.where(self.mapTemporary.map_image_rgb >
                                             (100, 100, 100), (255, 255, 255), (0, 0, 0)),
                                    dtype=np.uint8)[:, :, 0]
        straightened_out_lines = []
        search_angle = np.pi / 4
        snap_line_angle_delta_rad = self.args.snap_line_angle_delta_deg * np.pi / 180
        # Corrects line angles, position and extends lines
        for idx_line, line in enumerate(lines):
            # check and fix angle
            if line.angle > 0:
                deltas = []
                for i in range(4):
                    deltas.append(line.angle - search_angle * i)
                closest_angle = search_angle * np.argmin(np.abs(deltas))
            else:
                deltas = []
                for i in range(4):
                    deltas.append(line.angle + search_angle * i)
                closest_angle = -search_angle * np.argmin(np.abs(deltas))
            delta_angle = np.min(np.abs(deltas))

            # if angle delta to closest snap angle more than defined constant skip this line
            # Need to find the best value, this changes results a lot
            if abs(delta_angle) > snap_line_angle_delta_rad:
                continue
            # rotate line to closest angle
            if abs(delta_angle) > 0:
                line = self.controllerLine.rotate_line(line, closest_angle)

            # check and fix position
            # shift line to sides and chekc how many pixels covered
            shift_value = int(self.args.max_shift_after_rotate * self.args.pix_per_meter / 2)
            values = []
            for shift in range(-shift_value, shift_value):
                tmp_line = self.controllerLine.side_shift_line(line, shift)
                tmp_map = np.zeros_like(self.mapTemporary.map_image_greyscale)
                tmp_map = cv2.line(tmp_map,
                                   self.controllerPoint.to_int_list(tmp_line.start_point),
                                   self.controllerPoint.to_int_list(tmp_line.end_point),
                                   255,
                                   int(self.args.wall_width * self.args.pix_per_meter))
                covered_pixels = np.sum(np.logical_and(tmp_map, map_for_coverage))
                values.append(covered_pixels)
            # shift to max covered
            line = self.controllerLine.side_shift_line(line, range(-shift_value, shift_value)[np.argmax(values)])

            # extend line
            for is_end in (False, True):
                scores = []
                step_idx = 0
                stopping_param = 0
                while True:
                    tmp_line = copy.deepcopy(line)
                    tmp_line = self.controllerLine.extend_line(tmp_line, self.args.line_extend_step * self.args.pix_per_meter * step_idx,
                                                    is_end)
                    tmp_map = np.zeros_like(self.mapTemporary.map_image_greyscale)
                    tmp_map = cv2.line(tmp_map,
                                       self.controllerPoint.to_int_list(tmp_line.start_point),
                                       self.controllerPoint.to_int_list(tmp_line.end_point),
                                       255,
                                       int(self.args.wall_width * self.args.pix_per_meter))
                    covered_pixels = np.sum(np.logical_and(tmp_map, self.mapTemporary.map_image_rgb[:, :, 0]))
                    if scores:
                        if covered_pixels - scores[-1] > 10:  # smarter constant maybe from percent from new added area
                            stopping_param = 0
                        else:
                            stopping_param += 1

                    if stopping_param * self.args.line_extend_step >= self.args.max_wall_gap:
                        line = self.controllerLine.extend_line(line,
                            self.args.line_extend_step * self.args.pix_per_meter * (step_idx - stopping_param), is_end)
                        break
                    scores.append(covered_pixels)
                    step_idx += 1

            diagonal_line_angles = abs(np.array([0.7, 2.3]) - abs(line.angle))
            if (diagonal_line_angles > 0).any() and (
                    diagonal_line_angles <= 0.5).any() and self.controllerLine.get_lenght(line) < self.args.min_diagonal_line_length:
                logger.info(f'diagonal line is too short. SKIPPING')
                continue

            straightened_out_lines.append(line)

        # if self.args.debug_draw:
        #     self._debug_draw(self.mapTemporary.map_image_rgb, straightened_out_lines, 'straightened_out_lines')

        # new method
        lines_filtered = []
        for idx_line, line in enumerate(straightened_out_lines):
            is_too_close = False
            for idx_second_line in range(len(lines_filtered)):
                drone_width_pixels = self.args.drone_width * self.args.pix_per_meter
                delta_threshold = self.args.min_line_distance * self.args.pix_per_meter

                angle_delta = abs(line.angle - lines_filtered[idx_second_line].angle)
                if angle_delta < 0.01:
                    line_distance = ControllerLine.to_linestring(line).distance(ControllerLine.to_linestring(lines_filtered[idx_second_line]))
                    if line_distance < drone_width_pixels:
                        is_too_close = True
                        break
                # old method only for lines that are not parallel
                else:
                    # Euclidean distance between the line's starting and ending points
                    start_point_delta = np.sqrt(
                        (line.start_point.x - lines_filtered[idx_second_line].start_point.x) ** 2 +
                        (line.start_point.y - lines_filtered[idx_second_line].start_point.y) ** 2)
                    end_point_delta = np.sqrt(
                        (line.end_point.x - lines_filtered[idx_second_line].end_point.x) ** 2 +
                        (line.end_point.y - lines_filtered[idx_second_line].end_point.y) ** 2)

                    if start_point_delta < delta_threshold and end_point_delta < delta_threshold:
                        # lines position
                        tmp_mask = np.zeros_like(self.mapTemporary.map_image_greyscale)
                        tmp_line = copy.deepcopy(lines_filtered[idx_second_line])
                        tmp_line = self.controllerLine.extend_line(tmp_line, delta_threshold, True)
                        tmp_line = self.controllerLine.extend_line(tmp_line, delta_threshold, False)
                        tmp_mask = cv2.line(tmp_mask,
                                            self.controllerPoint.to_int_list(tmp_line.start_point),
                                            self.controllerPoint.to_int_list(tmp_line.end_point),
                                            255,
                                            int(2 * delta_threshold))
                        if tmp_mask[self.controllerPoint.to_int_list(line.start_point)[1], self.controllerPoint.to_int_list(line.start_point)[0]] and \
                                tmp_mask[self.controllerPoint.to_int_list(line.end_point)[1], self.controllerPoint.to_int_list(line.end_point)[0]]:
                            logger.info(f'line too close to other line. SKIPPING')
                            is_too_close = True
                            break
            if is_too_close:
                continue

            # check covered pixel density
            tmp_map = np.zeros_like(self.mapTemporary.map_image_greyscale)
            tmp_map = cv2.line(tmp_map,
                               self.controllerPoint.to_int_list(line.start_point),
                               self.controllerPoint.to_int_list(line.end_point),
                               255,
                               int(self.args.wall_width * self.args.pix_per_meter))
            covered_pixels = np.sum(np.logical_and(tmp_map, map_for_coverage))
            pixel_density = covered_pixels / self.controllerLine.get_lenght(line) / int(self.args.wall_width * self.args.pix_per_meter)
            if pixel_density < self.args.min_pixel_density:
                logger.info(f'Line pixel density is {pixel_density}. SKIPPING')
                continue

            diagonal_line_angles = abs(np.array([0.7, 2.3]) - abs(line.angle))
            if (diagonal_line_angles > 0).any() and (diagonal_line_angles <= 0.5).any():
                lines_crossed = 0

                # Utility functions
                # Checks if point is on the line segment p and q
                def on_segment(point_p, point_q, point_r):
                    if ((point_q.x <= max(point_p.x, point_r.x)) and (point_q.x >= min(point_p.x, point_r.x)) and
                            (point_q.y <= max(point_p.y, point_r.y)) and (point_q.y >= min(point_p.y, point_r.y))):
                        return True
                    return False

                # Calculates the orientation of the triplet points
                def orientation(point_p, point_q, point_r):
                    # 0 - collinear points
                    # 1 - clockwise points
                    # 2 - counterclockwise
                    val = (float(point_q.y - point_p.y) * (point_r.x - point_q.x)) - (
                                float(point_q.x - point_p.x) * (point_r.y - point_q.y))
                    if val > 0:
                        return 1
                    elif val < 0:
                        return 2
                    else:
                        return 0

                for idx_second_line in range(len(lines_filtered)):
                    # Checks if two lines intersect based on their orientation
                    start_p1 = line.start_point
                    end_p1 = line.end_point
                    start_p2 = lines_filtered[idx_second_line].start_point
                    end_p2 = lines_filtered[idx_second_line].end_point

                    o1 = orientation(start_p1, end_p1, start_p2)
                    o2 = orientation(start_p1, end_p1, end_p2)
                    o3 = orientation(start_p2, end_p2, start_p1)
                    o4 = orientation(start_p2, end_p2, end_p1)

                    # General case
                    if (o1 != o2) and (o3 != o4):
                        lines_crossed += 1

                    # Special Cases
                    elif (o1 == 0) and on_segment(start_p1, start_p2, end_p1):
                        lines_crossed += 1

                    elif (o2 == 0) and on_segment(start_p1, end_p2, end_p1):
                        lines_crossed += 1

                    elif (o3 == 0) and on_segment(start_p2, start_p1, end_p2):
                        lines_crossed += 1

                    elif (o4 == 0) and on_segment(start_p2, end_p1, end_p2):
                        lines_crossed += 1

                if lines_crossed > 1:
                    logger.info(f'Lines crossed {lines_crossed}. SKIPPING')
                    continue

            lines_filtered.append(line)

        # if self.args.debug_draw:
        #     self._debug_draw(self.mapTemporary.map_image_rgb, lines_filtered, 'filtered lines')

        # Check where lines cross and remove loose ends
        is_intersections = True
        min_line_len_pix =  self.args.min_line_length * self.args.pix_per_meter
        skip_pairs = []
        while is_intersections:
            is_intersections = False
            # order lines to start with shorter lines
            lines_lengths = [ControllerLine.get_lenght(i) for i in lines_filtered]
            lines_filtered = list(np.array(lines_filtered)[np.argsort(lines_lengths)])
            for first_idx, line in enumerate(lines_filtered):
                for second_idx, second_line in enumerate(lines_filtered):
                    is_skip = False
                    for pair in skip_pairs:
                        if first_idx in pair and second_idx in pair:
                            is_skip = True
                    if is_skip:
                        continue
                    if line != second_line:
                        # convert lines to linestrings
                        line_linestring = ControllerLine.to_linestring(line)
                        second_line_linestring = ControllerLine.to_linestring(second_line)
                        # get lines intersection
                        intersection = line_linestring.intersection(second_line_linestring)
                        if intersection:
                            # split line on intersection point
                            split_linestrings = list(split(line_linestring, intersection))
                            # len can be 1 if line ends at intersection point
                            if len(split_linestrings) > 1:
                                is_intersections = True
                                keep_split_lines = [False] * len(split_linestrings)
                                for split_idx, split_linestring in enumerate(split_linestrings):
                                    split_line = ControllerLine.from_linestring(split_linestring)
                                    # check line length
                                    line_len = ControllerLine.get_lenght(split_line)
                                    if line_len > min_line_len_pix:
                                        keep_split_lines[split_idx] = True

                                # after intersection
                                # remove original line and add split lines that are longer than defined minimum
                                if any(keep_split_lines) and not all(keep_split_lines):
                                    lines_filtered.remove(line)
                                    for split_idx, split_linestring in enumerate(split_linestrings):
                                        if keep_split_lines[split_idx]:
                                            lines_filtered.append(ControllerLine.from_linestring(split_linestring))
                                elif sum(keep_split_lines) == 0:
                                    lines_filtered.remove(line)
                                else:
                                    # keep original lines
                                    skip_pairs.append((first_idx, second_idx))
                                break
                if is_intersections:
                    break

        # extend lines to connect corners
        is_extend_lines = True
        while is_extend_lines:
            is_extend_lines = False
            # order lines to start with longer lines
            lines_lengths = [ControllerLine.get_lenght(i) for i in lines_filtered]
            lines_filtered = list(np.array(lines_filtered)[np.argsort(lines_lengths)][::-1])
            for line in lines_filtered:
                for second_idx, second_line in enumerate(lines_filtered):
                    # skip same line and lines with same angle
                    if line != second_line and abs(line.angle - second_line.angle) > 0.01:
                        # convert lines to linestrings
                        line_linestring = ControllerLine.to_linestring(line)
                        second_line_linestring = ControllerLine.to_linestring(second_line)
                        intersection = line_linestring.intersection(second_line_linestring)
                        # only check lines that dont intersect
                        if not intersection:
                            # iterate line endpoints
                            for line_endpoint in line_linestring.boundary:
                                for second_line_endpoint in second_line_linestring.boundary:
                                    # self._debug_draw(self.mapTemporary.map_image_rgb, [line, second_line], 'Loose ends removed', thinckness=2)
                                    distance = line_endpoint.distance(second_line_endpoint)
                                    if distance < self.args.max_wall_gap * self.args.pix_per_meter:
                                        is_extend_lines = True
                                        # extend lines algorithm
                                        # extend both lines
                                        line_long = ControllerLine.extend_line(line, 2 * distance, True)
                                        line_long = ControllerLine.extend_line(line_long, 2 * distance, False)
                                        line_long_linestring = ControllerLine.to_linestring(line_long)
                                        second_line_long = ControllerLine.extend_line(second_line, 2 * distance, True)
                                        second_line_long = ControllerLine.extend_line(second_line_long, 2 * distance, False)
                                        second_line_long_linestring = ControllerLine.to_linestring(second_line_long)
                                        intersection_point = line_long_linestring.intersection(second_line_long_linestring)
                                        # create new version of line
                                        # find point to extend
                                        if ControllerPoint.to_shapely_point(line.start_point) == line_endpoint:
                                            new_line = ControllerLine.create_line(ControllerPoint.from_shapely_point(intersection_point), line.end_point)
                                        else:
                                            new_line = ControllerLine.create_line(line.start_point, ControllerPoint.from_shapely_point(intersection_point))
                                        # create new version of second line
                                        # find point to extend
                                        if ControllerPoint.to_shapely_point(second_line.start_point) == second_line_endpoint:
                                            new_second_line = ControllerLine.create_line(ControllerPoint.from_shapely_point(intersection_point), second_line.end_point)
                                        else:
                                            new_second_line = ControllerLine.create_line(second_line.start_point, ControllerPoint.from_shapely_point(intersection_point))
                                        # edit filtered lines list
                                        lines_filtered.remove(line)
                                        lines_filtered.remove(second_line)
                                        lines_filtered.append(new_second_line)
                                        lines_filtered.append(new_line)
                                        break
                                if is_extend_lines:
                                    break
                    if is_extend_lines:
                        break
                if is_extend_lines:
                    break

        # # connect close parallel lines
        # is_connect_lines = True
        # lines_connected = []
        # while is_connect_lines:
        #     is_connect_lines = False
        #     # order lines to start with longer lines
        #     lines_lengths = [ControllerLine.get_lenght(i) for i in lines_filtered]
        #     lines_filtered = list(np.array(lines_filtered)[np.argsort(lines_lengths)][::-1])
        #     for first_idx, line in enumerate(lines_filtered):
        #         for second_idx, second_line in enumerate(lines_filtered):
        #             is_skip = False
        #             for pair in lines_connected:
        #                 if first_idx in pair and second_idx in pair:
        #                     is_skip = True
        #             if is_skip:
        #                 continue
        #             # skip same line and lines that are not parallel
        #             if line != second_line and abs(line.angle - second_line.angle) < 0.01:
        #                 # convert lines to linestrings
        #                 line_linestring = ControllerLine.to_linestring(line)
        #                 second_line_linestring = ControllerLine.to_linestring(second_line)
        #                 intersection = line_linestring.intersection(second_line_linestring)
        #                 # only check lines that dont intersect
        #                 if not intersection:
        #                     # iterate line endpoints
        #                     for line_endpoint in line_linestring.boundary:
        #                         for second_line_endpoint in second_line_linestring.boundary:
        #                             # self._debug_draw(self.mapTemporary.map_image_rgb, [line, second_line], 'Loose ends removed', thinckness=2)
        #                             distance = line_endpoint.distance(second_line_endpoint)
        #                             if distance < self.args.parallel_join_distance * self.args.pix_per_meter:
        #                                 is_connect_lines = True
        #                                 new_line = ControllerLine.create_line(
        #                                     ControllerPoint.from_shapely_point(line_endpoint),
        #                                     ControllerPoint.from_shapely_point(second_line_endpoint))
        #                                 lines_filtered.append(new_line)
        #                                 lines_connected.append((first_idx, second_idx))
        #                                 break
        #                         if is_connect_lines:
        #                             break
        #             if is_connect_lines:
        #                 break
        #         if is_connect_lines:
        #             break

        if self.args.debug_draw:
            self._debug_draw(self.mapTemporary.map_image_rgb, lines_filtered, 'processing finished', thinckness=2)


        for line in lines_filtered:
            self.mapTemporary.lines.append(line)
            # check position collision

            # draw line in final map
            self.mapTemporary.map_image_greyscale = cv2.line(self.mapTemporary.map_image_greyscale,
                                                             self.controllerPoint.to_int_list(line.start_point),
                                                             self.controllerPoint.to_int_list(line.end_point),
                                                             255,
                                                             int(self.args.wall_width * self.args.pix_per_meter))

    @property
    def is_path_points(self):
        return np.any(self.mapTemporary.path_points_in_meters.points)

    def split_line(self, line, inter_point, distance):
        drone_width = self.args.drone_width * self.args.pix_per_meter - distance
        tmp_line_1 = ControllerLine.create_line(inter_point, line.end_point)
        tmp_line_2 = ControllerLine.create_line(line.start_point, inter_point)
        tmp_line_1 = self.controllerLine.shorten_line(tmp_line_1, drone_width, cut_end=False)
        tmp_line_2 = self.controllerLine.shorten_line(tmp_line_2, drone_width, cut_end=True)
        new_lines = []
        for new_line in [tmp_line_1, tmp_line_2]:
            # check if line exists
            if new_line is not None:
                # check line length
                if self.controllerLine.get_lenght(new_line) > self.args.min_line_length * self.args.pix_per_meter:
                    # check covered pixel density
                    tmp_map = np.zeros_like(self.mapTemporary.map_image_greyscale)
                    tmp_map = cv2.line(tmp_map,
                                       self.controllerPoint.to_int_list(new_line.start_point),
                                       self.controllerPoint.to_int_list(new_line.end_point),
                                       255,
                                       int(self.args.wall_width * self.args.pix_per_meter))
                    covered_pixels = np.sum(np.logical_and(tmp_map, np.array(
                        np.where(self.mapTemporary.map_image_rgb > (100, 100, 100), (255, 255, 255), (0, 0, 0)), dtype=np.uint8)[:, :,
                                                                    0]))
                    pixel_density = covered_pixels / self.controllerLine.get_lenght(new_line) / int(
                        self.args.wall_width * self.args.pix_per_meter)
                    if pixel_density > self.args.min_pixel_density:
                        new_lines.append(new_line)
        return new_lines

    def process_path(self):
        if self.is_path_points:
            path_point_list = []
            for point in self.mapTemporary.path_points_in_meters.points:
                point_as_list = self.controllerPoint.to_list(point)
                point_pixels = np.matmul(self.mapTemporary.transformation_matrix_meters_to_pix,
                                         np.array([point_as_list[0], point_as_list[1], 1]))
                path_point_list.append(list(point_pixels[:2]))
            path_line_string = LineString(path_point_list)

            new_lines = []
            drone_width = self.args.drone_width * self.args.pix_per_meter
            for line in self.mapTemporary.lines:
                tmp_lines = [copy.deepcopy(line)]
                # check distances
                while any([drone_width - path_line_string.distance(self.controllerLine.to_linestring(l)) > 0.1 for l in tmp_lines]):
                    for idx, l in enumerate(tmp_lines):
                        distance = path_line_string.distance(self.controllerLine.to_linestring(l))
                        if drone_width - distance > 0.1:
                            p_nearest = nearest_points(path_line_string, self.controllerLine.to_linestring(l))
                            closest_val = np.inf
                            closest_idx = 0
                            for p_idx, p in enumerate(p_nearest):
                                p_distance = self.controllerLine.to_linestring(l).distance(p)
                                if p_distance < closest_val:
                                    closest_val = p_distance
                                    closest_idx = p_idx

                            inter_point = Point(p_nearest[closest_idx].x, p_nearest[closest_idx].y)
                            splitted_lines = self.split_line(l, inter_point, distance)
                            del tmp_lines[idx]
                            tmp_lines.extend(splitted_lines)
                            break

                new_lines.extend(tmp_lines)
            self.mapTemporary.lines = new_lines
            self.mapTemporary.map_image_greyscale = np.zeros_like(self.mapTemporary.map_image_greyscale)
            for line in self.mapTemporary.lines:
                self.mapTemporary.map_image_greyscale = cv2.line(self.mapTemporary.map_image_greyscale,
                                                                 self.controllerPoint.to_int_list(line.start_point),
                                                                 self.controllerPoint.to_int_list(line.end_point),
                                                                 255,
                                                                 int(self.args.wall_width * self.args.pix_per_meter))

    def rotate_and_scale_data(self):
        line_start_points_out = np.zeros((len(self.mapTemporary.lines), 3))
        line_end_points_out = np.zeros((len(self.mapTemporary.lines), 3))
        if np.any(self.mapTemporary.lines):
            for idx in range(0, len(self.mapTemporary.lines)):
                line_start_points_out[idx] = np.array([self.mapTemporary.lines[idx].start_point.x,
                                                       self.mapTemporary.lines[idx].start_point.y, 1])
                line_end_points_out[idx] = np.array([self.mapTemporary.lines[idx].end_point.x,
                                                     self.mapTemporary.lines[idx].end_point.y, 1])
        self.mapTemporary.line_start_points_in_pix = line_start_points_out
        self.mapTemporary.line_end_points_in_pix = line_end_points_out

        for idx in range(len(self.mapTemporary.point_cloud_in_meters.points)):
            result = np.matmul(self.mapTemporary.transformation_matrix_meters_to_pix,
                               np.array([self.mapTemporary.point_cloud_in_meters.points[idx].x,
                                         self.mapTemporary.point_cloud_in_meters.points[idx].y, 1]))
            self.mapTemporary.point_cloud_in_pix.points[idx].x = result[0]
            self.mapTemporary.point_cloud_in_pix.points[idx].y = result[1]
            self.mapTemporary.point_cloud_in_pix.points[idx].z = result[2]

        self.mapTemporary.path_points_in_pix = copy.deepcopy(self.mapTemporary.path_points_in_meters)
        for idx in range(len(self.mapTemporary.path_points_in_meters.points)):
            result = np.matmul(self.mapTemporary.transformation_matrix_meters_to_pix,
                               np.array([self.mapTemporary.path_points_in_meters.points[idx].x,
                                         self.mapTemporary.path_points_in_meters.points[idx].y, 1]))
            self.mapTemporary.path_points_in_pix.points[idx].x = result[0]
            self.mapTemporary.path_points_in_pix.points[idx].y = result[1]
            self.mapTemporary.path_points_in_pix.points[idx].z = result[2]

        self.mapTemporary.filtered_point_cloud_in_pix = copy.deepcopy(self.mapTemporary.filtered_point_cloud_in_meters)
        for idx in range(len(self.mapTemporary.filtered_point_cloud_in_meters.points)):
            result = np.matmul(self.mapTemporary.transformation_matrix_meters_to_pix,
                               np.array([self.mapTemporary.filtered_point_cloud_in_meters.points[idx].x,
                                         self.mapTemporary.filtered_point_cloud_in_meters.points[idx].y, 1]))
            self.mapTemporary.filtered_point_cloud_in_pix.points[idx].x = result[0]
            self.mapTemporary.filtered_point_cloud_in_pix.points[idx].y = result[1]
            self.mapTemporary.filtered_point_cloud_in_pix.points[idx].z = result[2]

if __name__ == '__main__':
    filepath_ply = r'C:\Users\vecin\Documents\PycharmProjects\asya\alarm_flight_ui_2022_Q1\test_data\master_DonMadden_P1_20_D_1_1_1.ply'
    filepath_txt = r'C:\Users\vecin\Documents\PycharmProjects\asya\alarm_flight_ui_2022_Q1\test_data\poses_DonMadden_M1_20_D_1_1_1.txt'
    controllerMap = ControllerMap(filepath_ply, filepath_txt, use_45deg_lines=False)
    # todo check for problems in line end cutting
    # todo extent lines to form a polygon
    # todo dont snap all lines to closest angles (if angle 12-32 just remove)

    a = 0