import copy
import os
import pickle
import time
from importlib import import_module
from typing import List, Union, Dict
from loguru import logger

from models.EnumTopicGroup import EnumTopicGroup
from models.LogRosTopic import LogRosTopic
from models.RosTopic import RosTopic
from models.RosTopics import RosTopics

IS_ROS = False
try:
    from std_msgs.msg import Int32, String
    from rospy.msg import AnyMsg
    import genpy
    import rosgraph
    import rostopic
    import bagpy
    import rosbag
    import rospy

    IS_ROS = True
except Exception as e:
    pass


class ControllerRosTopics:
    __instance = None

    def __init__(self):
        if ControllerRosTopics.__instance is not None:
            raise Exception("Cannot have 2 instances of singleton ControllerRosTopics")
        self.ros_topics_subscribed_full_paths_by_subscribers: Dict[str, any] = {}
        self.listener_ros_topics = None
        self.ros_pkg_to_class = {}

    def subscribe_to_all_topics(self, ros_topics):
        if IS_ROS:
            ros_topics_flattened = ControllerRosTopics.get_ros_topics_flattened(ros_topics)
            for ros_topic in ros_topics_flattened:
                if len(ros_topic.children_topics) == 0:
                    # only subscribe to the lowest level topics
                    if ros_topic.full_path not in self.ros_topics_subscribed_full_paths_by_subscribers.keys():
                        subscriber = rospy.Subscriber(ros_topic.full_path, AnyMsg, self.on_ros_topic_message)
                        self.ros_topics_subscribed_full_paths_by_subscribers[ros_topic.full_path] = subscriber

    def on_ros_topic_message(self, data):
        try:
            msg_topic = data._connection_header['topic']
            connection_header = data._connection_header['type'].split('/')
            ros_pkg = connection_header[0] + '.msg'
            msg_type = connection_header[1]
            msg_class = self.ros_pkg_to_class.setdefault(
                ros_pkg,
                getattr(import_module(ros_pkg), msg_type)
            )
            msg = ''
            buff = ''
            try:
                # some msg types don't have the _buff attribute attached!
                if hasattr(data, '_buff'):
                    buff = data._buff.hex()
                    msg = msg_class().deserialize(data._buff)
                else:
                    msg = str(data)
            except Exception as e:
                pass
                # logger.error("ERROR: Could not deserialize message for topic: " + msg_topic)
                # logger.exception(e)
            if self.listener_ros_topics is not None:
                self.listener_ros_topics(
                    msg_topic,
                    data._connection_header['type'],
                    str(msg),
                    buff,
                    data._connection_header)
        except Exception as e:
            #logger.exception(e)
            pass

    @staticmethod
    def instance():
        if ControllerRosTopics.__instance is None:
            ControllerRosTopics.__instance = ControllerRosTopics()
        return ControllerRosTopics.__instance

    @staticmethod
    def get_ros_topics_flattened(ros_topics: RosTopics) -> List[RosTopic]:
        topics = []
        if hasattr(ros_topics, 'ros_topics'):
            children = copy.deepcopy(ros_topics.ros_topics)
        else:
            children = ros_topics.children_topics
        for topic in children:
            topics.append(topic)
            topics += ControllerRosTopics.get_ros_topics_flattened(topic)
        # already contains all of children in root level
        for topic in topics:
            topic.children_topics = []
        return topics

    @staticmethod
    def get_ros_topics_tree(ros_topics_list: List[RosTopic]) -> RosTopics:
        ros_topics = RosTopics()
        ros_topics_list = sorted(ros_topics_list, key=lambda x: x.full_path)
        for topic in ros_topics_list:
            if '/' not in topic.full_path[1:]:
                ros_topics.ros_topics.append(topic)
            else:
                existing_topic = ControllerRosTopics.get_ros_topic_by_path(ros_topics, topic.full_path)
                if existing_topic is None:
                    full_path_parent = topic.full_path[:topic.full_path.rfind('/')]
                    parent_topic = ControllerRosTopics.get_ros_topic_by_path(ros_topics, full_path_parent)
                    if parent_topic is not None:
                        parent_topic.children_topics.append(topic)
                    else:
                        logger.error(f'Could not find parent topic for {topic.full_path}')

        return ros_topics

    @staticmethod
    def get_ros_topic_by_path(ros_topics: RosTopics, full_path: str):
        result = None
        for ros_topic in ros_topics.ros_topics:
            result = ControllerRosTopics.get_ros_topic_by_path_children(ros_topic, full_path)
            if result is not None:
                break
        return result

    @staticmethod
    def get_ros_topics_selected_from_all_top_topics(ros_topics: RosTopics, is_selected: bool) -> List[RosTopic]:
        result = []
        for ros_topic in ros_topics.ros_topics:
            if ros_topic.is_selected == is_selected:
                result.append(ros_topic)
            for child_ros_topic in ros_topic.children_topics:
                result += ControllerRosTopics.get_ros_topics_selected(child_ros_topic, is_selected)
        return result

    @staticmethod
    def get_ros_topics_selected(ros_topic: RosTopic, is_selected: bool) -> List[RosTopic]:
        result = []
        if ros_topic.is_selected == is_selected:
            result.append(ros_topic)
        for child_ros_topic in ros_topic.children_topics:
            result += ControllerRosTopics.get_ros_topics_selected(child_ros_topic, is_selected)
        return result


    @staticmethod
    def get_ros_topic_by_path_children(ros_topic: RosTopic, full_path: str):
        result = None
        if ros_topic.full_path == full_path:
            result = ros_topic
        else:
            for child_ros_topic in ros_topic.children_topics:
                result = ControllerRosTopics.get_ros_topic_by_path_children(child_ros_topic, full_path)
                if result is not None:
                    break
        return result

    @staticmethod
    def get_ros_topics():
        ros_topics = RosTopics()
        try:
            if IS_ROS:
                master = rosgraph.Master('/rostopic')
                published, subscribed = rostopic.get_topic_list(master=master)
            else:
                published = [('/fci/fly_to_poi/goal', 'adc_msgs/ADCFlyToPOIActionGoal', ['/gui']), ('/fci/fly_to_poi/cancel', 'actionlib_msgs/GoalID', ['/gui']), ('/fci/land/cancel', 'actionlib_msgs/GoalID', ['/gui']), ('/fci/takeoff/cancel', 'actionlib_msgs/GoalID', ['/gui']), ('/fci/takeoff/goal', 'adc_msgs/ADCTakeoffActionGoal', ['/gui']), ('/rosout', 'rosgraph_msgs/Log', ['/gui']), ('/rosout_agg', 'rosgraph_msgs/Log', ['/rosout']), ('/fci/yaw/cancel', 'actionlib_msgs/GoalID', ['/gui']), ('/fci/yaw/goal', 'adc_msgs/ADCYawActionGoal', ['/gui']), ('/fci/land/goal', 'adc_msgs/ADCLandActionGoal', ['/gui'])]
                subscribed = [('/fci/takeoff/feedback', 'adc_msgs/ADCTakeoffActionFeedback', ['/gui']), ('/fci/yaw/feedback', 'adc_msgs/ADCYawActionFeedback', ['/gui']), ('/fci/fly_to_poi/status', 'actionlib_msgs/GoalStatusArray', ['/gui']), ('/fci/yaw/status', 'actionlib_msgs/GoalStatusArray', ['/gui']), ('/fci/land/feedback', 'adc_msgs/ADCLandActionFeedback', ['/gui']), ('/fci/takeoff/status', 'actionlib_msgs/GoalStatusArray', ['/gui']), ('/fci/takeoff/result', 'adc_msgs/ADCTakeoffActionResult', ['/gui']), ('/fci/land/result', 'adc_msgs/ADCLandActionResult', ['/gui']), ('/rosout', 'rosgraph_msgs/Log', ['/rosout']), ('/fci/land/status', 'actionlib_msgs/GoalStatusArray', ['/gui']), ('/fci/fly_to_poi/feedback', 'adc_msgs/ADCFlyToPOIActionFeedback', ['/gui']), ('/rovio/odometry', 'nav_msgs/Odometry', ['/gui']), ('/fci/yaw/result', 'adc_msgs/ADCYawActionResult', ['/gui']), ('/fci/fly_to_poi/result', 'adc_msgs/ADCFlyToPOIActionResult', ['/gui'])]
            additional_alarm_topics = [
                ('/fci/path', 'nav_msgs/Path', ['gui']),
                ('/fci/setpoint_raw', 'geometry_msgs/PoseStamped', ['gui']),
                ('/fci/setpoint_raw_vel', 'geometry_msgs/TwistStamped', ['gui']),
                ('/mavros/estimator_status', 'adc_msgs/ADC_EstimatorStatus', ['gui']),
                ('/mavros/lidar/lidar_bottom', 'sensor_msgs/Range', ['gui']),
                ('/mavros/lidar/lidar_front', 'sensor_msgs/Range', ['gui']),
                ('/mavros/lidar/lidar_left', 'sensor_msgs/Range', ['gui']),
                ('/mavros/lidar/lidar_right', 'sensor_msgs/Range', ['gui']),
                ('/mavros/lidar/lidar_top', 'sensor_msgs/Range', ['gui']),
                ('/mavros/local_position/odom', 'nav_msgs/Odometry', ['gui']),
                ('/mavros/local_position/pose', 'geometry_msgs/PoseStamped', ['gui']),
                ('/mavros/local_position/velocity_body', 'geometry_msgs/TwistStamped', ['gui']),
                ('/mavros/state', 'mavros_msgs/State', ['gui']),
                ('/mavros/timesync_status', 'mavros_msgs/TimesyncStatus', ['gui']),
                ('/navigator_node/navmesh', 'visualization_msgs/MarkerArray', ['gui']),
                ('/rovio/VioHealth', 'adc_msgs/VioHealth', ['gui']),
                ('/rovio/odometry/uart', 'nav_msgs/Odometry', ['gui']),
                ('/rtabmap/info', 'rtabmap_ros/Info', ['gui']),
                ('/rtabmap/localization_pose', 'geometry_msgs/PoseWithCovarianceStamped', ['gui']),
                ('/rtabmap/rgbd_image_feature', 'rtabmap_ros/RGBDImage', ['gui']),
                ('/tf', 'tf2_msgs/TFMessage', ['gui']),
                ('/tf_static', 'tf2_msgs/TFMessage', ['gui'])
            ]

            topic_full_paths = []
            for topic_group_list in [published, subscribed, additional_alarm_topics]:
                for topic_full_path, topic_message_type, source_type in topic_group_list:
                    if topic_full_path not in topic_full_paths:
                        ros_topic = RosTopic()
                        ros_topic.name = os.path.basename(topic_full_path)
                        ros_topic.full_path = topic_full_path
                        ros_topic.message_type = topic_message_type
                        if topic_group_list == subscribed:
                            ros_topic.group = EnumTopicGroup.subscribed
                        else:
                            ros_topic.group = EnumTopicGroup.published
                        ros_topic.source_type = source_type
                        topic_full_paths.append(topic_full_path)

                        parts_of_paths = topic_full_path.split('/')[1:]
                        reconstructed_full_path = ''
                        ros_topic_parent: Union[RosTopic, None] = None
                        for part_of_path in parts_of_paths:
                            reconstructed_full_path += '/' + part_of_path
                            ros_topic_candidate = ControllerRosTopics.get_ros_topic_by_path(ros_topics, reconstructed_full_path)
                            if ros_topic_candidate is None:
                                ros_topic_candidate = RosTopic.from_dict(ros_topic.to_dict())
                                ros_topic_candidate.full_path = reconstructed_full_path
                                ros_topic_candidate.name = os.path.basename(reconstructed_full_path)
                                if ros_topic_parent is None:
                                    ros_topics.ros_topics.append(ros_topic_candidate)
                                else:
                                    ros_topic_parent.children_topics.append(ros_topic_candidate)
                            ros_topic_parent = ros_topic_candidate
            # ControllerRosTopics.write_to_bag_file(["/tf_static"])
            # ControllerRosTopics.read_bag_file(topics=['/fci/yaw/feedback'])
        except Exception as e:
            logger.error("Exception occured in get_ros_topics")
            logger.exception(e)
        return ros_topics

    @staticmethod
    def write_to_bag_file(logs: List[LogRosTopic], filepath: str):
        bag = None
        try:
            from controllers.ControllerChromium import ControllerChromium
            ControllerChromium.instance().add_log_info("UI", "Writing to bag file: {}".format(filepath))
            bag = rosbag.Bag(filepath, 'w')
            failed = 0
            for log in logs:
                try:
                    try:
                        data = AnyMsg()
                        if log.buff_base64:
                            data._buff = bytes.fromhex(log.buff_base64)
                        else:
                            data._buff = b''
                        data._connection_header = log.header
                        bag.write(
                            log.topic_full_path,
                            data,
                            genpy.Time.from_sec(log.timestamp_unix),
                            connection_header=log.header
                        )
                    except Exception as e:
                        if failed == 0:
                            logger.error("First Exception occured in write_to_bag_file inner")
                            logger.exception(e)
                        failed += 1
                        continue
                except Exception as e:
                    if failed == 0:
                        logger.error("First Exception occured in write_to_bag_file")
                        logger.exception(e)
                    failed += 1
            ControllerChromium.instance().add_log_info("UI", "Bag file writing finished. Successfully wrote {} out of {} topics".format(len(logs)-failed, len(logs)))
        except Exception as ex:
            logger.exception(ex)
            print(type(ex), ": Could not write to bag file")
            ControllerChromium.instance().add_log_info("UI", "Bag file writing failed")
        if bag is not None:
            bag.close()

    @staticmethod
    def read_bag_file(topics):
        try:
            # (a, b, c) = rostopic.get_topic_type('/fci/land/goal')
            # msg = rospy.wait_for_message(topics[0], "adc_msgs/ADCYawActionFeedback")
            # print(type(a))
            # print(type(b))
            bag = rosbag.Bag(os.path.join(os.path.dirname(__file__), 'test_bag  .bag'))
            for topic, msg, t in bag.read_messages(topics=topics):
                print("topic: ", topic)
                print(msg)
                break
        except Exception as ex:
            print(type(ex), ": Cannot read bag file")